//! Output of `Graph::parse`

use std::fmt;
use std::ops::{Deref, DerefMut};
use std::iter::IntoIterator;
use std::collections::{HashMap, HashSet};

use iri::Iri;
use object::Object;
use subject::Subject;

/// A set of Triples
#[derive(Clone, Debug, Default)]
pub struct Triples(pub Vec<Triple>);

fn map_blanks_values(triple: &Triple) -> Option<(Iri, Object)> {
    if triple.0.is_blank_node() {
        Some((triple.1.clone(), triple.2.clone()))
    } else {
        None
    }
}

impl Triples {
    /// Determines if two graphs are isomorphic.
    pub fn is_isomorphic(&self, other: &Self) -> bool {
        let blanks = self.iter().filter_map(map_blanks_values).collect::<HashSet<_>>();
        let other_blanks = other.iter().filter_map(map_blanks_values).collect::<HashSet<_>>();

        let non_blank_equality = self.iter()
            .filter(|&v| !v.0.is_blank_node())
            .all(|v| other.contains(v));

        non_blank_equality && blanks == other_blanks
    }

    fn create_map(&self) -> HashMap<Subject, (Iri, Object)> {
        let mut m = HashMap::new();

        for triple in &self.0 {
            let triple = triple.clone();
            m.insert(triple.0, (triple.1, triple.2));
        }

        m
    }
}

impl IntoIterator for Triples {
    type Item = Triple;
    type IntoIter = ::std::vec::IntoIter<Self::Item>;

    fn into_iter(self) -> Self::IntoIter {
        self.0.into_iter()
    }
}

impl Deref for Triples {
    type Target = Vec<Triple>;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl DerefMut for Triples {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.0
    }
}

impl fmt::Display for Triples {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {

        for triple in &self.0 {
            write!(f, "{}\n", triple)?;
        }

        Ok(())
    }
}

/// A single triple containing a subject, predicate, and object.
#[derive(Clone, Debug, PartialEq)]
pub struct Triple(pub Subject, pub Iri, pub Object);

impl fmt::Display for Triple {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{} {} {} .", self.0, self.1, self.2)
    }
}
